import os
import sys
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
from tqdm.notebook import tqdm
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.distributions.multivariate_normal import MultivariateNormal
from torch.distributions import MixtureSameFamily, Categorical
import json

from src.datasets.dist_for_ref import MixtureModel
from src.models.sbm import ScoreBaseModel, pretrain_sbm

from datetime import datetime

import tqdm

def get_log_dir():
    now = datetime.now()
    date_str = now.strftime("%Y-%m-%d_%H-%M-%S")  
    log_dir = os.path.join('logs/pretrain', f'pretrain_{date_str}')
    
    return log_dir

def plot_loss(df_res, output_dir):
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)

    index_list = []
    min_list = []
    index_list.append(0)
    for i in range(0, len(df_res)):
        index_list.append(i+1)
        min_list.append(min(df_res['loss'][:i+1]))

    df_res['min'] = min_list
    ax = df_res['min'].plot(color="blue", label='loss')

    fs = 15

    ax.set_xlabel('epoch', fontsize=fs)
    ax.set_ylabel('loss', fontsize=fs)
    # ax.set_yscale('log')
    ax.set_xscale('log')
    ax.set_ylim(0, df_res['loss'][0]+0.2)
    output_path = os.path.join(output_dir, 'pretrain.png')
    plt.savefig(output_path)
    # close the plot
    plt.close()

def pretrain_score():
    # fix seed
    torch.manual_seed(1234)

    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

    # Create the model
    mm = MixtureModel(config_dir="configs", device=device)
    print("mixture model created.")

    # Sample from the distribution
    samples = mm.sample(1000)

    # Calculate log probabilities
    log_probs = mm.log_prob(samples)

    with open("configs/pretrain_sbm.json", "r") as f:
        config_sbm = json.load(f)
    input_dim = config_sbm["input_dim"]
    lr = config_sbm["lr"]
    batch_size = config_sbm["batch_size"] # 2000
    num_epoch = config_sbm["num_epoch"] # 10000
    T = config_sbm["T"] # 10

    sbm = ScoreBaseModel(input_dim).to(device)
    optimizer = torch.optim.Adam(sbm.parameters(), lr=lr)
    # Loss
    criterion = torch.nn.MSELoss()

    df_res = pd.read_csv('')
    if df_res.empty:
        df_res = pretrain_sbm(batch_size, num_epoch, T, mm.dist, sbm, optimizer, criterion, device, log_dir=get_log_dir())

    now = datetime.now()
    now_str = now.strftime('%Y%m%d_%H%M%S')
    save_dir = os.path.join('outputs', 'pretrain', now_str)

    os.makedirs(save_dir, exist_ok=True)
    plot_loss(df_res, save_dir)

    torch.save(sbm.state_dict(), os.path.join(save_dir, 'score_base_model.pth'))

if __name__ == "__main__":
    pretrain_score()