import torch
import gc

def compute_conditional_expectation(model, model_potential, x_t, t, num_steps=100, T=10, num_samples=100, device='cuda'):
    step_size = t / num_steps
    batch_size = x_t.shape[0]

    # eval
    model.eval()
    model_potential.eval()

    def model_exp(x):
        return torch.exp(-model_potential(x))

    # x_t (shape: [batch_size * num_samples, input_dim])
    x_t_repeated = x_t.repeat_interleave(num_samples, dim=0).to(device) # (x_t[0], x_t[0], ... x_t[0], x_t[1], x_t[1], ..., x_t[1], ...)
    assert (x_t_repeated[0][0] == x_t_repeated[num_samples-1][0]) & (x_t_repeated[0][1] == x_t_repeated[num_samples-1][1])

    for i in range(num_steps):
        with torch.no_grad():
            t_tmp = t - t / num_steps * i
            if t_tmp > 0:
                batch_t = torch.ones((batch_size * num_samples, 1)).to(device) * t_tmp
                score = model(torch.cat([x_t_repeated, batch_t], axis=1))
                if i < num_steps - 1:
                    noise = torch.randn_like(x_t_repeated).to(device)
                else:
                    noise = torch.tensor(0).to(device)
                    noise = noise.repeat(batch_size * num_samples, 2)
                x_t_repeated = x_t_repeated + (x_t_repeated + 2 * score) * step_size + torch.sqrt(2 * torch.tensor(step_size)) * noise

                del batch_t, score, noise
                gc.collect()
                torch.cuda.empty_cache()

    del x_t
    gc.collect()
    torch.cuda.empty_cache()

    expectations = model_exp(x_t_repeated).squeeze(1)  # (batch_size * num_samples, )

    del x_t_repeated
    gc.collect()
    torch.cuda.empty_cache()

    expectations = expectations.view(batch_size, num_samples)  # (batch_size, num_samples)
    expectations = torch.mean(expectations, dim=1)  # (batch_size, )

    return expectations

def doob_h(model, model_potential, x, t, step_size):
    """
        x: requires_grad = True, (input_dim,)
    """
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    num_samples = 300 #0

    model_potential.eval()

    x = x.requires_grad_(True).to(device)
    model.train()

    x = x.requires_grad_(True).unsqueeze(0)

    t_gpu = torch.ones((x.shape[0], 1), requires_grad = False, dtype=x.dtype).to(device) * t

    score = model(torch.cat([x.float(), t_gpu], axis=1)) ## concat x and t

    #############################################################
    ########## THIS PART CAN BE REPLACED WITH jacobian ##########
    #############################################################
    gradients_first = torch.autograd.grad(score[0, 0], x, retain_graph=True)[0][0]
    gradients_second = torch.autograd.grad(score[0, 1], x)[0][0]

    model.eval()

    # Convert step_size to a tensor
    step_size_tensor = torch.tensor(step_size, device=device)

    mean = torch.exp(step_size_tensor) * x + (-2 * score) * (1-torch.exp(-step_size_tensor)) ## torch.tensor([?,?])
    std = torch.sqrt(torch.exp(2 * step_size_tensor)-1)

    x_t = torch.randn(num_samples, model.input_dim).to(device) * std + mean

    deriv_mean_x_by_x = torch.exp(step_size_tensor) + (-2 * gradients_first[0]) * (1-torch.exp(-step_size_tensor))
    deriv_mean_x_by_y = (-2 * gradients_first[1]) * (1-torch.exp(-step_size_tensor))
    deriv_mean_y_by_x = (-2 * gradients_second[0]) * (1-torch.exp(-step_size_tensor))
    deriv_mean_y_by_y = torch.exp(step_size_tensor) + (-2 * gradients_second[1]) * (1-torch.exp(-step_size_tensor))

    del gradients_first, gradients_second
    gc.collect()
    torch.cuda.empty_cache()

    mean_diff_x = deriv_mean_x_by_x * (x_t[:, 0] - mean[0][0]) / std**2 \
                + deriv_mean_y_by_x * (x_t[:, 1] - mean[0][1]) / std**2
    mean_diff_y = deriv_mean_x_by_y * (x_t[:, 0] - mean[0][0]) / std**2 \
                + deriv_mean_y_by_y * (x_t[:, 1] - mean[0][1]) / std**2

    # expectations = torch.stack([compute_conditional_expectation(model, exp_potential, x_t[i], max(t-step_size,0)) for i in range(x_t.shape[0])])
    t_e = max(t-step_size,0)
    expectations = compute_conditional_expectation(model, model_potential, x_t, t_e, num_steps=t_e*10)
    # print(expectations)

    sum_x = torch.sum(mean_diff_x * expectations)
    sum_y = torch.sum(mean_diff_y * expectations)

    denom = torch.sum(expectations)

    del deriv_mean_x_by_x, deriv_mean_x_by_y, deriv_mean_y_by_x, deriv_mean_y_by_y, expectations, x_t
    gc.collect()
    torch.cuda.empty_cache()

    eps = 1e-10

    return sum_x / (denom + eps), sum_y / (denom + eps)

def doob_h_tensorized(model, model_potential, x, t, step_size):
    """
    x: (batch_size, input_dim)
    """
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    num_samples = 300 #0
    model_potential.eval()
    model.train()

    input_dim = 2

    x = x.requires_grad_(True).to(device)
    # x = x.unsqueeze(0)

    t_gpu = torch.ones((x.shape[0], 1), requires_grad = False, dtype=x.dtype).to(device) * t
    # t_gpu = t_gpu.squeeze(1)


    score = model(torch.cat([x.float(), t_gpu], axis=1)) ## concat x and t

    # gradients = torch.autograd.grad(score, x, retain_graph=True)[0] # (batch_size, input_dim)
    
    gradients_first = []
    gradients_second = []
    for i in range(score.size(0)):
        grad_outputs = torch.ones_like(score[i][0])
        grad = torch.autograd.grad(score[i][0], x, grad_outputs=grad_outputs, retain_graph=True)[0][0]
        gradients_first.append(grad)
        grad = torch.autograd.grad(score[i][1], x, grad_outputs=grad_outputs, retain_graph=True)[0][0]
        gradients_second.append(grad)
    gradients_first = torch.stack(gradients_first).detach()
    gradients_second = torch.stack(gradients_second).detach()

    score = score.detach()

    model.eval()

    # Convert step_size to a tensor
    step_size_tensor = torch.tensor(step_size, device=device)

    mean = torch.exp(step_size_tensor) * x + (-2 * score) * (1-torch.exp(-step_size_tensor)) ## torch.tensor([?,?])
    std = torch.sqrt(torch.exp(2 * step_size_tensor)-1)

    ## x_t: (batch_size, num_samples, input_dim)
    x_t_original = torch.randn(x.shape[0], num_samples, x.shape[1]).to(device) * std + mean.unsqueeze(1)

    # x_t: (batch_size * num_samples, input_dim)
    x_t_reshaped = x_t_original.view(x.shape[0] * num_samples, input_dim)

    expectations = compute_conditional_expectation(model, model_potential, x_t_reshaped, max(t-step_size,0))

    deriv_mean_x_by_x = torch.exp(step_size_tensor) + (-2 * gradients_first[:,0]) * (1-torch.exp(-step_size_tensor))
    deriv_mean_x_by_y = (-2 * gradients_first[:,1]) * (1-torch.exp(-step_size_tensor))
    deriv_mean_y_by_x = (-2 * gradients_second[:,0]) * (1-torch.exp(-step_size_tensor))
    deriv_mean_y_by_y = torch.exp(step_size_tensor) + (-2 * gradients_second[:,1]) * (1-torch.exp(-step_size_tensor))

    del gradients_first, gradients_second
    gc.collect()
    torch.cuda.empty_cache()

    x_t = x_t_reshaped.view(x.shape[0], num_samples, input_dim)
    
    deriv_mean_x_by_x = deriv_mean_x_by_x.unsqueeze(1).repeat(1, num_samples)
    deriv_mean_x_by_y = deriv_mean_x_by_y.unsqueeze(1).repeat(1, num_samples)
    deriv_mean_y_by_x = deriv_mean_y_by_x.unsqueeze(1).repeat(1, num_samples)
    deriv_mean_y_by_y = deriv_mean_y_by_y.unsqueeze(1).repeat(1, num_samples)

    mean = mean.unsqueeze(1).repeat(1, x_t.shape[1], 1)

    mean_diff_x = deriv_mean_x_by_x * (x_t[:, :, 0] - mean[:, :, 0]) / std**2 \
                + deriv_mean_y_by_x * (x_t[:, :, 1] - mean[:, :, 1]) / std**2
    mean_diff_y = deriv_mean_x_by_y * (x_t[:, :, 0] - mean[:, :, 0]) / std**2 \
                + deriv_mean_y_by_y * (x_t[:, :, 1] - mean[:, :, 1]) / std**2

    expectations = expectations.reshape(x.shape[0], num_samples)  # (batch_size, num_samples)

    sum_x = torch.sum(mean_diff_x * expectations, dim=1)
    sum_y = torch.sum(mean_diff_y * expectations, dim=1)

    denom = torch.sum(expectations, dim=1)

    del deriv_mean_x_by_x, deriv_mean_y_by_y, expectations, x_t, score
    gc.collect()
    torch.cuda.empty_cache()

    eps = 1e-10

    return sum_x / (denom + eps), sum_y / (denom + eps)

