import torch
import torch.nn as nn
import torch.nn.functional as F
import pandas as pd
import numpy as np
from tqdm import tqdm

import wandb

class ScoreBaseModel(nn.Module):
    def __init__(self, input_dim, mid_dim=64):
        super(ScoreBaseModel, self).__init__()
        self.input_dim = input_dim
        self.mid_dim = mid_dim
        self.fc1 = nn.Linear(input_dim+1, mid_dim) # dtype=torch.float32)
        self.fc2 = nn.Linear(mid_dim, mid_dim//2) # dtype=torch.float32)
        self.fc3 = nn.Linear(mid_dim//2, mid_dim//4) # dtype=torch.float32)
        self.fc4 = nn.Linear(mid_dim//4, input_dim) # dtype=torch.float32)
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

         # Move the model parameters to the device
        self.to(self.device)

        return

    def forward(self, x):
        x = F.tanh(self.fc1(x.float().to(self.device)))
        x = F.tanh(self.fc2(x))
        x = F.tanh(self.fc3(x))
        x = self.fc4(x)
        return x

def pretrain_sbm(batch_size, num_epoch, T, dist, model, optimizer, criterion, device, log_dir='logs'):
    wandb.init(project='mixture_model')
    np.random.seed(1234)
    torch.manual_seed(1234)
    out_list = []
    model.train()
    for epoch in tqdm(range(num_epoch), leave=False):
        sample = dist.sample((batch_size,))
        t = torch.full((batch_size,), np.random.rand() *T).to(device)
        t = t.unsqueeze(1)
        t_expand = t.expand(-1, 2)
        noise = torch.randn_like(sample).to(device)
        xt = torch.exp(-t_expand) * sample + torch.sqrt(1-torch.exp(-2*t_expand)) * noise
        x = torch.cat([xt, t], axis=1).to(device)
        pred_y = model(x)

        true_y = - noise / torch.sqrt(1-torch.exp(-2*t_expand))

        optimizer.zero_grad()
        loss = criterion(true_y, pred_y)
        loss.backward()
        optimizer.step()

        out_list.append([epoch, loss.item()])
        wandb.log({'loss': loss.item()})

        # print(f"Epoch [{epoch+1}/{num_epoch}], Loss: {loss.item()}")

    print("Finished Pretraining")

    df_res = pd.DataFrame(out_list, columns=['epoch','loss'])
    return df_res