import os
import sys
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
from tqdm import tqdm
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.distributions.multivariate_normal import MultivariateNormal
from torch.distributions import MixtureSameFamily, Categorical
import json
import datetime

import gc

import sys
sys.path.append("")
# print(sys.path)

from src.datasets.dist_for_ref import MixtureModel
from src.models.sbm import ScoreBaseModel, pretrain_sbm
from src.doob_h.doob_h import doob_h_tensorized # doob_h
from src.models.nn_potential import ApproxPotential
from src.utils.sampling import doob_langevin_monte_carlo_modified, new_sbm_based_langevin_monte_carlo

import wandb

def plot(x_doobs, x_refs, sbm_path, potential_paths, dir):
    # Create subplots
    len_x_doobs = x_doobs.shape[0]
    fig, axs = plt.subplots(1, len_x_doobs+1, figsize=(5*len_x_doobs+2, 5))
    # fig, axs = plt.subplots(2,2, figsize=(10,10))
    x_doob_np = x_doobs.cpu().numpy()
    x_ref_np = x_refs.cpu().numpy()
    # subplot
    # plt.subplots_adjust(wspace=0.25, hspace=0.25)

    n_bins = 30
    fs = 30

    # Plot x_ref
    axs[0].set_title('reference', fontsize=fs)
    axs[0].hist2d(
        x_ref_np[0,:,0],
        x_ref_np[0,:,1],
        range=((-5, 5), (-5, 5)),
        cmap='viridis',
        bins=n_bins,
    )
    axs[0].set_aspect('equal', adjustable='box')
    axs[0].set_xlim([-5, 5])
    axs[0].set_ylim([-5, 5])
    axs[0].set_xlabel('x1', fontsize=fs)
    axs[0].set_ylabel('x2', fontsize=fs)

    # Plot x_doob
    for i in range(len_x_doobs):
        axs[1+i].set_title('k = ' + str(i+1), fontsize=fs)
        axs[1+i].hist2d(
            x_doob_np[i,:,0],
            x_doob_np[i,:,1],
            range=((-5, 5), (-5, 5)),
            cmap='viridis',
            bins=n_bins,
        )
        axs[1+i].set_aspect('equal', adjustable='box')
        axs[1+i].set_xlim([-5, 5])
        axs[1+i].set_ylim([-5, 5])
        axs[1+i].set_xlabel('x1', fontsize=fs)
        axs[1+i].set_ylabel('x2', fontsize=fs)

    filename = os.path.join(dir, 'doob_h_monte_carlo.png')
    plt.savefig(filename)
    with open(os.path.join(dir, 'model_path.txt'), mode='w') as f:
        f.write(sbm_path + '\n')
        for potential_path in potential_paths:
            f.write(potential_path + '\n')

import os
import matplotlib.pyplot as plt
import seaborn as sns

def KDE_plot(x_doobs, x_refs, sbm_path, potential_paths, dir):
    # Create subplots
    len_x_doobs = x_doobs.shape[0]
    fig, axs = plt.subplots(1, len_x_doobs + 1, figsize=(5 * len_x_doobs + 4, 5))
    x_doob_np = x_doobs.cpu().numpy()
    x_ref_np = x_refs.cpu().numpy()

    # Adjust subplot spacing
    # plt.subplots_adjust(wspace=0.2, hspace=0.3)

    xmax = 5
    xmin = -5
    ymax = 5
    ymin = -5

    # fontsize
    fs = 30
    fs_axis = 15
    levels = 100

    # Plot x_ref with Kernel Density Estimation (KDE)
    axs[0].set_title('reference', fontsize=fs)
    sns.kdeplot(
        x=x_ref_np[0, :, 0], 
        y=x_ref_np[0, :, 1], 
        fill=True, 
        cmap='Blues', 
        ax=axs[0],
        levels=levels,
        clip=[[-5, 5], [-5, 5]]
    )
    axs[0].set_aspect('equal', adjustable='box')
    axs[0].set_xlim([-5, 5])
    axs[0].set_ylim([-5, 5])
    # x
    axs[0].set_xlabel('x1', fontsize=fs_axis)
    # y
    axs[0].set_ylabel('x2', fontsize=fs_axis)

    # Plot x_doob with Kernel Density Estimation (KDE)
    for i in range(len_x_doobs):
        axs[1 + i].set_title('k = ' + str(i + 1), fontsize=fs)
        sns.kdeplot(
            x=x_doob_np[i, :, 0], 
            y=x_doob_np[i, :, 1], 
            fill=True, 
            cmap='Blues', 
            ax=axs[1 + i],
            levels=levels,
            clip=[[-5, 5], [-5, 5]]
        )
        axs[1 + i].set_aspect('equal', adjustable='box')
        axs[1 + i].set_xlim([-5, 5])
        axs[1 + i].set_ylim([-5, 5])
        # x
        axs[1 + i].set_xlabel('x1', fontsize=fs_axis)
        # y
        axs[1 + i].set_ylabel('x2', fontsize=fs_axis)

    # Save the plot as an image
    filename = os.path.join(dir, 'doob_h_monte_carlo_kde.png')
    plt.savefig(filename)

    # Save the model paths to a text file
    with open(os.path.join(dir, 'model_path.txt'), mode='w') as f:
        f.write(sbm_path + '\n')
        for potential_path in potential_paths:
            f.write(potential_path + '\n')


def main():
    # load config
    config_doob_h_path = 'configs/doob_h.json'
    with open(config_doob_h_path) as f:
        config_doob_h = json.load(f)

    wandb.init(
        project="mog_doob",
        config=config_doob_h
    )

    sampled_dir = config_doob_h["sampled_dir"]
    # date
    date = pd.Timestamp.now().strftime('%Y%m_%d%H%M%S')
    now = datetime.datetime.now()
    now_str = now.strftime('%Y%m%d_%H%M%S')
    if sampled_dir == None:
        dir = os.path.join('outputs', 'figures', 'doob_h_MC', now_str)
        os.makedirs(dir, exist_ok=True)
    else:
        dir = os.path.join('outputs', 'figures', 'doob_h_MC_plot_only', now_str)
        os.makedirs(dir, exist_ok=True)
    # dir, filename

    sbm_path = config_doob_h["sbm_path"]
    sbm = ScoreBaseModel(2)
    sbm.load_state_dict(torch.load(sbm_path))
    sbm.eval()
    sbm = nn.DataParallel(sbm)
    sbm.cuda()
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

    T = 10
    num_steps = 50
    num_repeat_samples = 100
    num_samples_i = 100

    x_doob = []
    x_ref = []

    print("### Start doob h ###")

    path_num = config_doob_h["path_num"]
    potential_paths = []
    for i in range(1, path_num+1):
        potential_paths.append(config_doob_h["potential_path_"+str(i)])

    x_doobs = []
    x_refs = []
    path_iter = 0

    if sampled_dir == None:
        for potential_path in potential_paths:
            print("### potential_path: ", potential_path, "###")
            x_doob = []
            x_ref = []
            model_potential = ApproxPotential(2)
            model_potential.load_state_dict(torch.load(potential_path))
            model_potential = nn.DataParallel(model_potential)
            model_potential.cuda()
            model_potential.eval()
            for i in tqdm(range(num_repeat_samples)):
                # x_doob_i, x_ref_i = doob_langevin_monte_carlo(sbm, model_potential, num_samples_i, num_steps, step_size)
                x_doob_i = doob_langevin_monte_carlo_modified(sbm, model_potential, num_samples_i, num_steps, T, device = device)
                x_doob.append(x_doob_i)
                if path_iter == 0:
                    x_ref_i = new_sbm_based_langevin_monte_carlo(sbm, num_samples_i, num_steps, T, device = device)
                    x_ref.append(x_ref_i)
            # x_doob to csv
            x_doob = torch.cat(x_doob, axis=0)
            # os.path.join(dir, 'doob_i.csv')
            x_doob_df = pd.DataFrame(x_doob.cpu().numpy())
            x_doob_df.to_csv(os.path.join(dir, 'doob'+str(path_iter)+'.csv'))
            x_doobs.append(x_doob)
            if path_iter == 0:
                x_ref = torch.cat(x_ref, axis=0)
                x_ref_df = pd.DataFrame(x_ref.cpu().numpy())
                x_ref_df.to_csv(os.path.join(dir, 'ref'+str(path_iter)+'.csv'))
                x_refs.append(x_ref)
            print("### Finished potential_path: ", potential_path, "###")
            print("x_doob: ", x_doob.shape)
            path_iter += 1
    else:
        # sampled_dir has doob0.csv, doob1.csv, ... 
        for i in range(path_num):
            doob_path = os.path.join(sampled_dir, 'doob'+str(i)+'.csv')
            print("### doob_path: ", doob_path, "###")
            x_doob_i = pd.read_csv(doob_path, header=None, index_col=0)
            x_doob_i = torch.tensor(x_doob_i.values)
            # length of doob_i
            print("x_doob_i: ", x_doob_i.shape)
            x_doobs.append(x_doob_i)
            if i == 0:
                ref_path = os.path.join(sampled_dir, 'ref'+str(i)+'.csv')
                x_ref_i = pd.read_csv(ref_path, header=None, index_col=0)
                x_ref_i = torch.tensor(x_ref_i.values)
                x_refs.append(x_ref_i)
            wandb.log({"i": i})
    
    x_doobs = torch.stack(x_doobs, axis=0)
    x_refs = torch.stack(x_refs, axis=0)
    plot(x_doobs, x_refs, sbm_path, potential_paths, dir)
    KDE_plot(x_doobs, x_refs, sbm_path, potential_paths, dir)

    print("### Finished doob h ###")

if __name__ == "__main__":
    main()